@@ -23,6 +23,11 @@ const ConfigSchema = require('../config-schema')
 
 const LocationSuffixRegExp = /(:\d+)(:\d+)?$/
 
+// Increment this when changing the serialization format of `${ATOM_HOME}/storage/application.json` used by
+// AtomApplication::saveCurrentWindowOptions() and AtomApplication::loadPreviousWindowOptions() in a backward-
+// incompatible way.
+const APPLICATION_STATE_VERSION = '1'
+
 const getDefaultPath = () => {
   const editor = atom.workspace.getActiveTextEditor()
   if (!editor || !editor.getPath()) {
@@ -34,6 +39,81 @@ const getDefaultPath = () => {
   }
 }
 
+const getSocketSecretPath = (atomVersion) => {
+  const {username} = os.userInfo()
+  const atomHome = path.resolve(process.env.ATOM_HOME)
+
+  return path.join(atomHome, `.atom-socket-secret-${username}-${atomVersion}`)
+}
+
+const getSocketPath = (socketSecret) => {
+  if (!socketSecret) {
+    return null
+  }
+
+  // Hash the secret to create the socket name to not expose it.
+  const socketName = crypto
+    .createHmac('sha256', socketSecret)
+    .update('socketName')
+    .digest('hex')
+    .substr(0, 12)
+
+  if (process.platform === 'win32') {
+    return `\\\\.\\pipe\\atom-${socketName}-sock`
+  } else {
+    return path.join(os.tmpdir(), `atom-${socketName}.sock`)
+  }
+}
+
+const getExistingSocketSecret = (atomVersion) => {
+  const socketSecretPath = getSocketSecretPath(atomVersion)
+
+  if (!fs.existsSync(socketSecretPath)) {
+    return null
+  }
+
+  return fs.readFileSync(socketSecretPath, 'utf8')
+}
+
+const createSocketSecret = (atomVersion) => {
+  const socketSecret = crypto.randomBytes(16).toString('hex')
+
+  fs.writeFileSync(getSocketSecretPath(atomVersion), socketSecret, {encoding: 'utf8', mode: 0o600})
+
+  return socketSecret
+}
+
+const encryptOptions = (options, secret) => {
+  const message = JSON.stringify(options)
+
+  const initVector = crypto.randomBytes(16)
+
+  const cipher = crypto.createCipheriv('aes-256-gcm', secret, initVector)
+
+  let content = cipher.update(message, 'utf8', 'hex')
+  content += cipher.final('hex')
+
+  const authTag = cipher.getAuthTag().toString('hex')
+
+  return JSON.stringify({
+    authTag,
+    content,
+    initVector: initVector.toString('hex')
+  })
+}
+
+const decryptOptions = (optionsMessage, secret) => {
+  const {authTag, content, initVector} = JSON.parse(optionsMessage)
+
+  const decipher = crypto.createDecipheriv('aes-256-gcm', secret, Buffer.from(initVector, 'hex'))
+  decipher.setAuthTag(Buffer.from(authTag, 'hex'))
+
+  let message = decipher.update(content, 'hex', 'utf8')
+  message += decipher.final('utf8')
+
+  return JSON.parse(message)
+}
+
 // The application's singleton class.
 //
 // It's the entry point into the Atom application and maintains the global state
@@ -43,56 +123,36 @@ module.exports =
 class AtomApplication extends EventEmitter {
   // Public: The entry point into the Atom application.
   static open (options) {
-    if (!options.socketPath) {
-      const {username} = os.userInfo()
-
-      // Lowercasing the ATOM_HOME to make sure that we don't get multiple sockets
-      // on case-insensitive filesystems due to arbitrary case differences in paths.
-      const atomHomeUnique = path.resolve(process.env.ATOM_HOME).toLowerCase()
-      const hash = crypto
-        .createHash('sha1')
-        .update(options.version)
-        .update('|')
-        .update(process.arch)
-        .update('|')
-        .update(username || '')
-        .update('|')
-        .update(atomHomeUnique)
-
-      // We only keep the first 12 characters of the hash as not to have excessively long
-      // socket file. Note that macOS/BSD limit the length of socket file paths (see #15081).
-      // The replace calls convert the digest into "URL and Filename Safe" encoding (see RFC 4648).
-      const atomInstanceDigest = hash
-        .digest('base64')
-        .substring(0, 12)
-        .replace(/\+/g, '-')
-        .replace(/\//g, '_')
-
-      if (process.platform === 'win32') {
-        options.socketPath = `\\\\.\\pipe\\atom-${atomInstanceDigest}-sock`
-      } else {
-        options.socketPath = path.join(os.tmpdir(), `atom-${atomInstanceDigest}.sock`)
-      }
-    }
+    const socketSecret = getExistingSocketSecret(options.version)
+    const socketPath = getSocketPath(socketSecret)
+    const createApplication = options.createApplication || (async () => {
+      const app = new AtomApplication(options)
+      await app.initialize(options)
+      return app
+    })
 
     // FIXME: Sometimes when socketPath doesn't exist, net.connect would strangely
     // take a few seconds to trigger 'error' event, it could be a bug of node
     // or electron, before it's fixed we check the existence of socketPath to
     // speedup startup.
-    if ((process.platform !== 'win32' && !fs.existsSync(options.socketPath)) ||
-        options.test || options.benchmark || options.benchmarkTest) {
-      new AtomApplication(options).initialize(options)
-      return
+    if (
+      !socketPath || options.test || options.benchmark || options.benchmarkTest ||
+      (process.platform !== 'win32' && !fs.existsSync(socketPath))
+    ) {
+      return createApplication(options)
     }
 
-    const client = net.connect({path: options.socketPath}, () => {
-      client.write(JSON.stringify(options), () => {
-        client.end()
-        app.quit()
+    return new Promise(resolve => {
+      const client = net.connect({path: socketPath}, () => {
+        client.write(encryptOptions(options, socketSecret), () => {
+          client.end()
+          app.quit()
+          resolve(null)
+        })
       })
-    })
 
-    client.on('error', () => new AtomApplication(options).initialize(options))
+      client.on('error', () => resolve(createApplication(options)))
+    })
   }
 
   exit (status) {
@@ -110,11 +170,14 @@ class AtomApplication extends EventEmitter {
     this.version = options.version
     this.devMode = options.devMode
     this.safeMode = options.safeMode
-    this.socketPath = options.socketPath
     this.logFile = options.logFile
     this.userDataDir = options.userDataDir
     this._killProcess = options.killProcess || process.kill.bind(process)
-    if (options.test || options.benchmark || options.benchmarkTest) this.socketPath = null
+
+    if (!options.test && !options.benchmark && !options.benchmarkTest) {
+      this.socketSecret = createSocketSecret(this.version)
+      this.socketPath = getSocketPath(this.socketSecret)
+    }
 
     this.waitSessionsByWindow = new Map()
     this.windowStack = new WindowStack()
@@ -189,33 +252,35 @@ class AtomApplication extends EventEmitter {
       this.config.onDidChange('core.colorProfile', () => this.promptForRestart())
     }
 
-    const optionsForWindowsToOpen = []
-
+    let optionsForWindowsToOpen = []
     let shouldReopenPreviousWindows = false
 
     if (options.test || options.benchmark || options.benchmarkTest) {
       optionsForWindowsToOpen.push(options)
+    } else if (options.newWindow) {
+      shouldReopenPreviousWindows = false
     } else if ((options.pathsToOpen && options.pathsToOpen.length > 0) ||
                (options.urlsToOpen && options.urlsToOpen.length > 0)) {
       optionsForWindowsToOpen.push(options)
       shouldReopenPreviousWindows = this.config.get('core.restorePreviousWindowsOnStart') === 'always'
-    } else if (options.newWindow) {
-      shouldReopenPreviousWindows = false
     } else {
       shouldReopenPreviousWindows = this.config.get('core.restorePreviousWindowsOnStart') !== 'no'
     }
 
     if (shouldReopenPreviousWindows) {
-      for (const previousOptions of await this.loadPreviousWindowOptions()) {
-        optionsForWindowsToOpen.push(Object.assign({}, options, previousOptions))
-      }
+      optionsForWindowsToOpen = [...await this.loadPreviousWindowOptions(), ...optionsForWindowsToOpen]
     }
 
     if (optionsForWindowsToOpen.length === 0) {
       optionsForWindowsToOpen.push(options)
     }
 
-    return optionsForWindowsToOpen.map(options => this.openWithOptions(options))
+    // Preserve window opening order
+    const windows = []
+    for (const options of optionsForWindowsToOpen) {
+      windows.push(await this.openWithOptions(options))
+    }
+    return windows
   }
 
   openWithOptions (options) {
@@ -236,10 +301,13 @@ class AtomApplication extends EventEmitter {
       timeout,
       clearWindowState,
       addToLastWindow,
+      preserveFocus,
       env
     } = options
 
-    app.focus()
+    if (!preserveFocus) {
+      app.focus()
+    }
 
     if (test) {
       return this.runTests({
@@ -276,11 +344,14 @@ class AtomApplication extends EventEmitter {
         addToLastWindow,
         env
       })
-    } else if (urlsToOpen.length > 0) {
-      return urlsToOpen.map(urlToOpen => this.openUrl({urlToOpen, devMode, safeMode, env}))
+    } else if (urlsToOpen && urlsToOpen.length > 0) {
+      return Promise.all(
+        urlsToOpen.map(urlToOpen => this.openUrl({urlToOpen, devMode, safeMode, env}))
+      )
     } else {
-      // Always open a editor window if this is the first instance of Atom.
+      // Always open an editor window if this is the first instance of Atom.
       return this.openPath({
+        pathToOpen: null,
         pidToKillWhenClosed,
         newWindow,
         devMode,
@@ -293,6 +364,11 @@ class AtomApplication extends EventEmitter {
     }
   }
 
+  // Public: Create a new {AtomWindow} bound to this application.
+  createWindow (settings) {
+    return new AtomWindow(this, this.fileRecoveryService, settings)
+  }
+
   // Public: Removes the {AtomWindow} from the global window list.
   removeWindow (window) {
     this.windowStack.removeWindow(window)
@@ -328,6 +404,7 @@ class AtomApplication extends EventEmitter {
         window.browserWindow.removeListener('blur', blurHandler)
       })
       window.browserWindow.webContents.once('did-finish-load', blurHandler)
+      this.saveCurrentWindowOptions(false)
     }
   }
 
@@ -351,11 +428,21 @@ class AtomApplication extends EventEmitter {
     const server = net.createServer(connection => {
       let data = ''
       connection.on('data', chunk => { data += chunk })
-      connection.on('end', () => this.openWithOptions(JSON.parse(data)))
+      connection.on('end', () => {
+        try {
+          const options = decryptOptions(data, this.socketSecret)
+          this.openWithOptions(options)
+        } catch (e) {
+          // Error while parsing/decrypting the options passed by the client.
+          // We cannot trust the client, aborting.
+        }
+      })
     })
 
-    server.listen(this.socketPath)
-    server.on('error', error => console.error('Application server failed', error))
+    return new Promise(resolve => {
+      server.listen(this.socketPath, resolve)
+      server.on('error', error => console.error('Application server failed', error))
+    })
   }
 
   deleteSocketFile () {
@@ -373,19 +460,37 @@ class AtomApplication extends EventEmitter {
     }
   }
 
+  deleteSocketSecretFile () {
+    if (!this.socketSecret) {
+      return
+    }
+
+    const socketSecretPath = getSocketSecretPath(this.version)
+
+    if (fs.existsSync(socketSecretPath)) {
+      try {
+        fs.unlinkSync(socketSecretPath)
+      } catch (error) {
+        // Ignore ENOENT errors in case the file was deleted between the exists
+        // check and the call to unlink sync.
+        if (error.code !== 'ENOENT') throw error
+      }
+    }
+  }
+
   // Registers basic application commands, non-idempotent.
   handleEvents () {
-    const getLoadSettings = includeWindow => {
-      const window = this.focusedWindow()
+    const createOpenSettings = ({event, sameWindow}) => {
+      const targetWindow = event ? this.atomWindowForEvent(event) : this.focusedWindow()
       return {
-        devMode: window && window.devMode,
-        safeMode: window && window.safeMode,
-        window: includeWindow && window
+        devMode: targetWindow ? targetWindow.devMode : false,
+        safeMode: targetWindow ? targetWindow.safeMode : false,
+        window: sameWindow && targetWindow ? targetWindow : null
       }
     }
 
     this.on('application:quit', () => app.quit())
-    this.on('application:new-window', () => this.openPath(getLoadSettings(false)))
+    this.on('application:new-window', () => this.openPath(createOpenSettings({})))
     this.on('application:new-file', () => (this.focusedWindow() || this).openPath())
     this.on('application:open-dev', () => this.promptForPathToOpen('all', {devMode: true}))
     this.on('application:open-safe', () => this.promptForPathToOpen('all', {safeMode: true}))
@@ -414,9 +519,16 @@ class AtomApplication extends EventEmitter {
         this.openPaths({ pathsToOpen: paths })
       })
 
-      this.on('application:open', () => this.promptForPathToOpen('all', getLoadSettings(true), getDefaultPath()))
-      this.on('application:open-file', () => this.promptForPathToOpen('file', getLoadSettings(true), getDefaultPath()))
-      this.on('application:open-folder', () => this.promptForPathToOpen('folder', getLoadSettings(true), getDefaultPath()))
+      this.on('application:open', () => {
+        this.promptForPathToOpen('all', createOpenSettings({sameWindow: true}), getDefaultPath())
+      })
+      this.on('application:open-file', () => {
+        this.promptForPathToOpen('file', createOpenSettings({sameWindow: true}), getDefaultPath())
+      })
+      this.on('application:open-folder', () => {
+        this.promptForPathToOpen('folder', createOpenSettings({sameWindow: true}), getDefaultPath())
+      })
+
       this.on('application:bring-all-windows-to-front', () => Menu.sendActionToFirstResponder('arrangeInFront:'))
       this.on('application:hide', () => Menu.sendActionToFirstResponder('hide:'))
       this.on('application:hide-other-applications', () => Menu.sendActionToFirstResponder('hideOtherApplications:'))
@@ -484,8 +596,12 @@ class AtomApplication extends EventEmitter {
     this.disposable.add(ipcHelpers.on(app, 'will-quit', () => {
       this.killAllProcesses()
       this.deleteSocketFile()
+      this.deleteSocketSecretFile()
     }))
 
+    // Triggered by the 'open-file' event from Electron:
+    // https://electronjs.org/docs/api/app#event-open-file-macos
+    // For example, this is fired when a file is dragged and dropped onto the Atom application icon in the dock.
     this.disposable.add(ipcHelpers.on(app, 'open-file', (event, pathToOpen) => {
       event.preventDefault()
       this.openPath({pathToOpen})
@@ -519,25 +635,36 @@ class AtomApplication extends EventEmitter {
       }
     }))
 
-    // A request from the associated render process to open a new render process.
-    this.disposable.add(ipcHelpers.on(ipcMain, 'open', (event, options) => {
-      const window = this.atomWindowForEvent(event)
+    // A request from the associated render process to open a set of paths using the standard window location logic.
+    // Used for application:reopen-project.
+    this.disposable.add(ipcHelpers.on(ipcMain, 'open', (_event, options) => {
       if (options) {
         if (typeof options.pathsToOpen === 'string') {
           options.pathsToOpen = [options.pathsToOpen]
         }
 
         if (options.pathsToOpen && options.pathsToOpen.length > 0) {
-          options.window = window
           this.openPaths(options)
         } else {
-          this.addWindow(new AtomWindow(this, this.fileRecoveryService, options))
+          this.addWindow(this.createWindow(options))
         }
       } else {
         this.promptForPathToOpen('all', {window})
       }
     }))
 
+    // Prompt for a file, folder, or either, then open the chosen paths. Files will be opened in the originating
+    // window; folders will be opened in a new window unless an existing window exactly contains all of them.
+    this.disposable.add(ipcHelpers.on(ipcMain, 'open-chosen-any', (event, defaultPath) => {
+      this.promptForPathToOpen('all', createOpenSettings({event, sameWindow: true}), defaultPath)
+    }))
+    this.disposable.add(ipcHelpers.on(ipcMain, 'open-chosen-file', (event, defaultPath) => {
+      this.promptForPathToOpen('file', createOpenSettings({event, sameWindow: true}), defaultPath)
+    }))
+    this.disposable.add(ipcHelpers.on(ipcMain, 'open-chosen-folder', (event, defaultPath) => {
+      this.promptForPathToOpen('folder', createOpenSettings({event}), defaultPath)
+    }))
+
     this.disposable.add(ipcHelpers.on(ipcMain, 'update-application-menu', (event, template, menu) => {
       const window = BrowserWindow.fromWebContents(event.sender)
       if (this.applicationMenu) this.applicationMenu.update(window, template, menu)
@@ -564,22 +691,9 @@ class AtomApplication extends EventEmitter {
       this.emit(command)
     }))
 
-    this.disposable.add(ipcHelpers.on(ipcMain, 'open-command', (event, command, defaultPath) => {
-      switch (command) {
-        case 'application:open':
-          return this.promptForPathToOpen('all', getLoadSettings(true), defaultPath)
-        case 'application:open-file':
-          return this.promptForPathToOpen('file', getLoadSettings(true), defaultPath)
-        case 'application:open-folder':
-          return this.promptForPathToOpen('folder', getLoadSettings(true), defaultPath)
-        default:
-          return console.log(`Invalid open-command received: ${command}`)
-      }
-    }))
-
     this.disposable.add(ipcHelpers.on(ipcMain, 'window-command', (event, command, ...args) => {
       const window = BrowserWindow.fromWebContents(event.sender)
-      return window.emit(command, ...args)
+      return window && window.emit(command, ...args)
     }))
 
     this.disposable.add(ipcHelpers.respondTo('window-method', (browserWindow, method, ...args) => {
@@ -651,10 +765,6 @@ class AtomApplication extends EventEmitter {
       this.fileRecoveryService.didSavePath(window, path)
     ))
 
-    this.disposable.add(ipcHelpers.on(ipcMain, 'did-change-paths', () =>
-      this.saveCurrentWindowOptions(false)
-    ))
-
     this.disposable.add(this.disableZoomOnDisplayChange())
   }
 
@@ -753,10 +863,12 @@ class AtomApplication extends EventEmitter {
     })
   }
 
-  // Returns the {AtomWindow} for the given paths.
-  windowForPaths (pathsToOpen, devMode) {
-    return this.getAllWindows().find(window =>
-      window.devMode === devMode && window.containsPaths(pathsToOpen)
+  // Returns the {AtomWindow} for the given locations.
+  windowForLocations (locationsToOpen, devMode, safeMode) {
+    return this.getLastFocusedWindow(window =>
+      window.devMode === devMode &&
+      window.safeMode === safeMode &&
+      window.containsLocations(locationsToOpen)
     )
   }
 
@@ -846,7 +958,7 @@ class AtomApplication extends EventEmitter {
   //   :windowDimensions - Object with height and width keys.
   //   :window - {AtomWindow} to open file paths in.
   //   :addToLastWindow - Boolean of whether this should be opened in last focused window.
-  openPaths ({
+  async openPaths ({
     pathsToOpen,
     foldersToOpen,
     executedFrom,
@@ -869,20 +981,19 @@ class AtomApplication extends EventEmitter {
     safeMode = Boolean(safeMode)
     clearWindowState = Boolean(clearWindowState)
 
-    const locationsToOpen = pathsToOpen.map(pathToOpen => {
-      return this.parsePathToOpen(pathToOpen, executedFrom, {
-        forceAddToWindow: addToLastWindow,
+    const locationsToOpen = await Promise.all(
+      pathsToOpen.map(pathToOpen => this.parsePathToOpen(pathToOpen, executedFrom, {
         hasWaitSession: pidToKillWhenClosed != null
-      })
-    })
+      }))
+    )
 
     for (const folderToOpen of foldersToOpen) {
       locationsToOpen.push({
         pathToOpen: folderToOpen,
         initialLine: null,
         initialColumn: null,
-        mustBeDirectory: true,
-        forceAddToWindow: addToLastWindow,
+        exists: true,
+        isDirectory: true,
         hasWaitSession: pidToKillWhenClosed != null
       })
     }
@@ -891,27 +1002,42 @@ class AtomApplication extends EventEmitter {
       return
     }
 
-    const normalizedPathsToOpen = locationsToOpen.map(location => location.pathToOpen).filter(Boolean)
+    const hasNonEmptyPath = locationsToOpen.some(location => location.pathToOpen)
+    const createNewWindow = newWindow || !hasNonEmptyPath
 
     let existingWindow
 
-    // Explicitly provided AtomWindow has precedence unless a new window is forced.
-    if (!newWindow) {
+    if (!createNewWindow) {
+      // An explicitly provided AtomWindow has precedence.
       existingWindow = window
-    }
 
-    // If no window is specified, a new window is not forced, and at least one path is provided, locate
-    // an existing window that contains all paths.
-    if (!existingWindow && !newWindow && normalizedPathsToOpen.length > 0) {
-      existingWindow = this.windowForPaths(normalizedPathsToOpen, devMode)
-    }
+      // If no window is specified and at least one path is provided, locate an existing window that contains all
+      // provided paths.
+      if (!existingWindow && hasNonEmptyPath) {
+        existingWindow = this.windowForLocations(locationsToOpen, devMode, safeMode)
+      }
 
-    // No window specified, new window not forced, no existing window found, and addition to the last window
-    // requested. Find the last focused window.
-    if (!existingWindow && !newWindow && addToLastWindow) {
-      let lastWindow = window || this.getLastFocusedWindow()
-      if (lastWindow && lastWindow.devMode === devMode) {
-        existingWindow = lastWindow
+      // No window specified, no existing window found, and addition to the last window requested. Find the last
+      // focused window that matches the requested dev and safe modes.
+      if (!existingWindow && addToLastWindow) {
+        existingWindow = this.getLastFocusedWindow(win => {
+          return win.devMode === devMode && win.safeMode === safeMode
+        })
+      }
+
+      // Fall back to the last focused window that has no project roots.
+      if (!existingWindow) {
+        existingWindow = this.getLastFocusedWindow(win => !win.hasProjectPaths())
+      }
+
+      // One last case: if *no* paths are directories, add to the last focused window.
+      if (!existingWindow) {
+        const noDirectories = locationsToOpen.every(location => !location.isDirectory)
+        if (noDirectories) {
+          existingWindow = this.getLastFocusedWindow(win => {
+            return win.devMode === devMode && win.safeMode === safeMode
+          })
+        }
       }
     }
 
@@ -942,7 +1068,7 @@ class AtomApplication extends EventEmitter {
       if (!resourcePath) resourcePath = this.resourcePath
       if (!windowDimensions) windowDimensions = this.getDimensionsForNewWindow()
 
-      openedWindow = new AtomWindow(this, this.fileRecoveryService, {
+      openedWindow = this.createWindow({
         locationsToOpen,
         windowInitializationScript,
         resourcePath,
@@ -963,7 +1089,7 @@ class AtomApplication extends EventEmitter {
       }
       this.waitSessionsByWindow.get(openedWindow).push({
         pid: pidToKillWhenClosed,
-        remainingPaths: new Set(normalizedPathsToOpen)
+        remainingPaths: new Set(locationsToOpen.map(location => location.pathToOpen).filter(Boolean))
       })
     }
 
@@ -1015,28 +1141,58 @@ class AtomApplication extends EventEmitter {
   async saveCurrentWindowOptions (allowEmpty = false) {
     if (this.quitting) return
 
-    const states = []
-    for (let window of this.getAllWindows()) {
-      if (!window.isSpec) states.push({initialPaths: window.representedDirectoryPaths})
+    const state = {
+      version: APPLICATION_STATE_VERSION,
+      windows: this.getAllWindows()
+        .filter(window => !window.isSpec)
+        .map(window => ({projectRoots: window.projectRoots}))
     }
-    states.reverse()
+    state.windows.reverse()
 
-    if (states.length > 0 || allowEmpty) {
-      await this.storageFolder.store('application.json', states)
+    if (state.windows.length > 0 || allowEmpty) {
+      await this.storageFolder.store('application.json', state)
       this.emit('application:did-save-state')
     }
   }
 
   async loadPreviousWindowOptions () {
-    const states = await this.storageFolder.load('application.json')
-    if (states) {
-      return states.map(state => ({
-        foldersToOpen: state.initialPaths,
-        urlsToOpen: [],
+    const state = await this.storageFolder.load('application.json')
+    if (!state) {
+      return []
+    }
+
+    if (state.version === APPLICATION_STATE_VERSION) {
+      // Atom >=1.36.1
+      // Schema: {version: '1', windows: [{projectRoots: ['<root-dir>', ...]}, ...]}
+      return state.windows.map(each => ({
+        foldersToOpen: each.projectRoots,
         devMode: this.devMode,
-        safeMode: this.safeMode
+        safeMode: this.safeMode,
+        newWindow: true
       }))
+    } else if (state.version === undefined) {
+      // Atom <= 1.36.0
+      // Schema: [{initialPaths: ['<root-dir>', ...]}, ...]
+      return await Promise.all(
+        state.map(async windowState => {
+          // Classify each window's initialPaths as directories or non-directories
+          const classifiedPaths = await Promise.all(
+            windowState.initialPaths.map(initialPath => new Promise(resolve => {
+              fs.isDirectory(initialPath, isDir => resolve({initialPath, isDir}))
+            }))
+          )
+
+          // Only accept initialPaths that are existing directories
+          return {
+            foldersToOpen: classifiedPaths.filter(({isDir}) => isDir).map(({initialPath}) => initialPath),
+            devMode: this.devMode,
+            safeMode: this.safeMode,
+            newWindow: true
+          }
+        })
+      )
     } else {
+      // Unrecognized version (from a newer Atom?)
       return []
     }
   }
@@ -1083,6 +1239,7 @@ class AtomApplication extends EventEmitter {
     if (bestWindow) {
       bestWindow.sendURIMessage(url)
       bestWindow.focus()
+      return bestWindow
     } else {
       let windowInitializationScript
       let {resourcePath} = this
@@ -1100,7 +1257,7 @@ class AtomApplication extends EventEmitter {
       }
 
       const windowDimensions = this.getDimensionsForNewWindow()
-      const window = new AtomWindow(this, this.fileRecoveryService, {
+      const window = this.createWindow({
         resourcePath,
         windowInitializationScript,
         devMode,
@@ -1124,7 +1281,7 @@ class AtomApplication extends EventEmitter {
     const packagePath = this.getPackageManager(devMode).resolvePackagePath(packageName)
     const windowInitializationScript = path.resolve(packagePath, packageUrlMain)
     const windowDimensions = this.getDimensionsForNewWindow()
-    const window = new AtomWindow(this, this.fileRecoveryService, {
+    const window = this.createWindow({
       windowInitializationScript,
       resourcePath: this.resourcePath,
       devMode,
@@ -1206,7 +1363,7 @@ class AtomApplication extends EventEmitter {
     if (safeMode == null) {
       safeMode = false
     }
-    const window = new AtomWindow(this, this.fileRecoveryService, {
+    const window = this.createWindow({
       windowInitializationScript,
       resourcePath,
       headless,
@@ -1255,7 +1412,7 @@ class AtomApplication extends EventEmitter {
     const devMode = true
     const isSpec = true
     const safeMode = false
-    const window = new AtomWindow(this, this.fileRecoveryService, {
+    const window = this.createWindow({
       windowInitializationScript,
       resourcePath,
       headless,
@@ -1310,31 +1467,58 @@ class AtomApplication extends EventEmitter {
     }
   }
 
-  parsePathToOpen (pathToOpen, executedFrom, extra) {
-    let initialColumn, initialLine
+  async parsePathToOpen (pathToOpen, executedFrom, extra) {
+    const result = Object.assign({
+      pathToOpen,
+      initialColumn: null,
+      initialLine: null,
+      exists: false,
+      isDirectory: false,
+      isFile: false
+    }, extra)
+
     if (!pathToOpen) {
-      return {pathToOpen}
+      return result
     }
 
-    pathToOpen = pathToOpen.replace(/[:\s]+$/, '')
-    const match = pathToOpen.match(LocationSuffixRegExp)
+    result.pathToOpen = result.pathToOpen.replace(/[:\s]+$/, '')
+    const match = result.pathToOpen.match(LocationSuffixRegExp)
 
     if (match != null) {
-      pathToOpen = pathToOpen.slice(0, -match[0].length)
+      result.pathToOpen = result.pathToOpen.slice(0, -match[0].length)
       if (match[1]) {
-        initialLine = Math.max(0, parseInt(match[1].slice(1)) - 1)
+        result.initialLine = Math.max(0, parseInt(match[1].slice(1), 10) - 1)
       }
       if (match[2]) {
-        initialColumn = Math.max(0, parseInt(match[2].slice(1)) - 1)
+        result.initialColumn = Math.max(0, parseInt(match[2].slice(1), 10) - 1)
       }
-    } else {
-      initialLine = initialColumn = null
     }
 
-    const normalizedPath = path.normalize(path.resolve(executedFrom, fs.normalize(pathToOpen)))
-    if (!url.parse(pathToOpen).protocol) pathToOpen = normalizedPath
+    const normalizedPath = path.normalize(path.resolve(executedFrom, fs.normalize(result.pathToOpen)))
+    if (!url.parse(pathToOpen).protocol) {
+      result.pathToOpen = normalizedPath
+    }
 
-    return Object.assign({pathToOpen, initialLine, initialColumn}, extra)
+    await new Promise((resolve, reject) => {
+      fs.stat(result.pathToOpen, (err, st) => {
+        if (err) {
+          if (err.code === 'ENOENT' || err.code === 'EACCES') {
+            result.exists = false
+            resolve()
+          } else {
+            reject(err)
+          }
+          return
+        }
+
+        result.exists = true
+        result.isFile = st.isFile()
+        result.isDirectory = st.isDirectory()
+        resolve()
+      })
+    })
+
+    return result
   }
 
   // Opens a native dialog to prompt the user for a path.
@@ -1389,8 +1573,8 @@ class AtomApplication extends EventEmitter {
       }
     })()
 
-    // Show the open dialog as child window on Windows and Linux, and as
-    // independent dialog on macOS. This matches most native apps.
+    // Show the open dialog as child window on Windows and Linux, and as an independent dialog on macOS. This matches
+    // most native apps.
     const parentWindow = process.platform === 'darwin' ? null : BrowserWindow.getFocusedWindow()
 
     const openOptions = {
